# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2018 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Views for maintaining vendor catalogs
"""

from __future__ import unicode_literals, absolute_import

import logging

import six

from rattail.db import model, api
from rattail.vendors.catalogs import iter_catalog_parsers

import colander
from deform import widget as dfwidget
from webhelpers2.html import tags

from tailbone import forms
from tailbone.db import Session
from tailbone.views.batch import FileBatchMasterView


log = logging.getLogger(__name__)


class VendorCatalogsView(FileBatchMasterView):
    """
    Master view for vendor catalog batches.
    """
    model_class = model.VendorCatalog
    model_row_class = model.VendorCatalogRow
    default_handler_spec = 'rattail.batch.vendorcatalog:VendorCatalogHandler'
    url_prefix = '/vendors/catalogs'
    template_prefix = '/batch/vendorcatalog'
    editable = False
    rows_bulk_deletable = True

    grid_columns = [
        'id',
        'created',
        'created_by',
        'vendor',
        'effective',
        'filename',
        'rowcount',
        'executed',
    ]

    form_fields = [
        'id',
        'vendor',
        'filename',
        'future',
        'effective',
        'created',
        'created_by',
        'rowcount',
        'executed',
        'executed_by',
    ]

    row_grid_columns = [
        'sequence',
        'upc',
        'brand_name',
        'description',
        'size',
        'vendor_code',
        'old_unit_cost',
        'unit_cost',
        'unit_cost_diff',
        'starts',
        'status_code',
    ]

    row_form_fields = [
        'sequence',
        'item_entry',
        'product',
        'upc',
        'brand_name',
        'description',
        'size',
        'old_vendor_code',
        'vendor_code',
        'old_case_size',
        'case_size',
        'old_case_cost',
        'case_cost',
        'case_cost_diff',
        'old_unit_cost',
        'unit_cost',
        'unit_cost_diff',
        'suggested_retail',
        'starts',
        'ends',
        'discount_starts',
        'discount_ends',
        'discount_amount',
        'discount_percent',
        'status_code',
        'status_text',
    ]

    def get_parsers(self):
        if not hasattr(self, 'parsers'):
            self.parsers = sorted(iter_catalog_parsers(), key=lambda p: p.display)
        return self.parsers

    def configure_grid(self, g):
        super(VendorCatalogsView, self).configure_grid(g)
        g.joiners['vendor'] = lambda q: q.join(model.Vendor)
        g.filters['vendor'] = g.make_filter('vendor', model.Vendor.name,
                                            default_active=True, default_verb='contains')
        g.sorters['vendor'] = g.make_sorter(model.Vendor.name)

    def get_instance_title(self, batch):
        return six.text_type(batch.vendor)

    def configure_form(self, f):
        super(VendorCatalogsView, self).configure_form(f)

        # vendor
        f.set_renderer('vendor', self.render_vendor)
        if self.creating:
            f.replace('vendor', 'vendor_uuid')
            f.set_node('vendor_uuid', colander.String())
            vendor_display = ""
            if self.request.method == 'POST':
                if self.request.POST.get('vendor_uuid'):
                    vendor = self.Session.query(model.Vendor).get(self.request.POST['vendor_uuid'])
                    if vendor:
                        vendor_display = six.text_type(vendor)
            vendors_url = self.request.route_url('vendors.autocomplete')
            f.set_widget('vendor_uuid', forms.widgets.JQueryAutocompleteWidget(
                field_display=vendor_display, service_url=vendors_url))
            f.set_label('vendor_uuid', "Vendor")
        else:
            f.set_readonly('vendor')

        # filename
        f.set_label('filename', "Catalog File")

        if self.creating:

            f.set_fields([
                'filename',
                'parser_key',
                'vendor_uuid',
                'future',
            ])

            parser_values = [(p.key, p.display) for p in self.get_parsers()]
            parser_values.insert(0, ('', "(please choose)"))
            f.set_widget('parser_key', dfwidget.SelectWidget(values=parser_values))
            f.set_label('parser_key', "File Type")

        # effective
        if not self.creating:
            f.set_readonly('effective')

    def render_vendor(self, batch, field):
        vendor = batch.vendor
        if not vendor:
            return ""
        text = "({}) {}".format(vendor.id, vendor.name)
        url = self.request.route_url('vendors.view', uuid=vendor.uuid)
        return tags.link_to(text, url)

    def get_batch_kwargs(self, batch):
        kwargs = super(VendorCatalogsView, self).get_batch_kwargs(batch)
        kwargs['parser_key'] = batch.parser_key
        if batch.vendor:
            kwargs['vendor'] = batch.vendor
        elif batch.vendor_uuid:
            kwargs['vendor_uuid'] = batch.vendor_uuid
        kwargs['future'] = batch.future
        return kwargs

    def configure_row_grid(self, g):
        super(VendorCatalogsView, self).configure_row_grid(g)
        batch = self.get_instance()

        # starts
        if not batch.future:
            g.hide_column('starts')

        g.set_label('upc', "UPC")
        g.set_label('brand_name', "Brand")
        g.set_label('old_unit_cost', "Old Cost")
        g.set_label('unit_cost', "New Cost")
        g.set_label('unit_cost_diff', "Diff.")

    def row_grid_extra_class(self, row, i):
        if row.status_code == row.STATUS_PRODUCT_NOT_FOUND:
            return 'warning'
        if row.status_code in (row.STATUS_NEW_COST,
                               row.STATUS_UPDATE_COST, # TODO: deprecate/remove this one
                               row.STATUS_CHANGE_VENDOR_ITEM_CODE,
                               row.STATUS_CHANGE_CASE_SIZE,
                               row.STATUS_CHANGE_COST):
            return 'notice'

    def configure_row_form(self, f):
        super(VendorCatalogsView, self).configure_row_form(f)
        f.set_renderer('product', self.render_product)
        f.set_type('discount_percent', 'percent')

    def template_kwargs_create(self, **kwargs):
        parsers = self.get_parsers()
        for parser in parsers:
            if parser.vendor_key:
                vendor = api.get_vendor(Session(), parser.vendor_key)
                if vendor:
                    parser.vendormap_value = "{{uuid: '{}', name: '{}'}}".format(
                        vendor.uuid, vendor.name.replace("'", "\\'"))
                else:
                    log.warning("vendor '{}' not found for parser: {}".format(
                        parser.vendor_key, parser.key))
                    parser.vendormap_value = 'null'
            else:
                parser.vendormap_value = 'null'
        kwargs['parsers'] = parsers
        return kwargs


def includeme(config):
    VendorCatalogsView.defaults(config)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2019 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Employee Views
"""

from __future__ import unicode_literals, absolute_import

import six
import sqlalchemy as sa

from rattail.db import model

import colander
from deform import widget as dfwidget
from webhelpers2.html import tags, HTML

from tailbone import grids
from tailbone.db import Session
from tailbone.views import MasterView, AutocompleteView


class EmployeesView(MasterView):
    """
    Master view for the Employee class.
    """
    model_class = model.Employee
    has_versions = True

    labels = {
        'id': "ID",
    }

    grid_columns = [
        'id',
        'first_name',
        'last_name',
        'phone',
        'email',
        'status',
    ]

    form_fields = [
        'person',
        'first_name',
        'last_name',
        'display_name',
        'phone',
        'email',
        'status',
        'full_time',
        'full_time_start',
        'id',
        'stores',
        'departments',
    ]

    def configure_grid(self, g):
        super(EmployeesView, self).configure_grid(g)
        route_prefix = self.get_route_prefix()

        g.joiners['phone'] = lambda q: q.outerjoin(model.EmployeePhoneNumber, sa.and_(
            model.EmployeePhoneNumber.parent_uuid == model.Employee.uuid,
            model.EmployeePhoneNumber.preference == 1))
        g.joiners['email'] = lambda q: q.outerjoin(model.EmployeeEmailAddress, sa.and_(
            model.EmployeeEmailAddress.parent_uuid == model.Employee.uuid,
            model.EmployeeEmailAddress.preference == 1))

        g.filters['first_name'] = g.make_filter('first_name', model.Person.first_name)
        g.filters['last_name'] = g.make_filter('last_name', model.Person.last_name)

        g.filters['email'] = g.make_filter('email', model.EmployeeEmailAddress.address,
                                           label="Email Address")
        g.set_filter('phone', model.EmployeePhoneNumber.number,
                     label="Phone Number",
                     factory=grids.filters.AlchemyPhoneNumberFilter)

        # id
        if self.request.has_perm('{}.edit'.format(route_prefix)):
            g.set_link('id')
        else:
            g.hide_column('id')
            del g.filters['id']

        # status
        if self.request.has_perm('{}.edit'.format(route_prefix)):
            g.set_enum('status', self.enum.EMPLOYEE_STATUS)
            g.filters['status'].default_active = True
            g.filters['status'].default_verb = 'equal'
            # TODO: why must we set unicode string value here?
            g.filters['status'].default_value = six.text_type(self.enum.EMPLOYEE_STATUS_CURRENT)
        else:
            g.hide_column('status')
            del g.filters['status']

        g.filters['first_name'].default_active = True
        g.filters['first_name'].default_verb = 'contains'

        g.filters['last_name'].default_active = True
        g.filters['last_name'].default_verb = 'contains'

        g.sorters['first_name'] = lambda q, d: q.order_by(getattr(model.Person.first_name, d)())
        g.sorters['last_name'] = lambda q, d: q.order_by(getattr(model.Person.last_name, d)())

        g.sorters['email'] = lambda q, d: q.order_by(getattr(model.EmployeeEmailAddress.address, d)())
        g.sorters['phone'] = lambda q, d: q.order_by(getattr(model.EmployeePhoneNumber.number, d)())

        g.set_sort_defaults('first_name')

        g.set_label('phone', "Phone Number")
        g.set_label('email', "Email Address")

        g.set_link('first_name')
        g.set_link('last_name')

    def query(self, session):
        q = session.query(model.Employee).join(model.Person)
        if not self.request.has_perm('employees.edit'):
            q = q.filter(model.Employee.status == self.enum.EMPLOYEE_STATUS_CURRENT)
        return q

    def editable_instance(self, employee):
        if self.rattail_config.demo():
            return not bool(employee.user and employee.user.username == 'chuck')
        return True

    def deletable_instance(self, employee):
        if self.rattail_config.demo():
            return not bool(employee.user and employee.user.username == 'chuck')
        return True

    def configure_form(self, f):
        super(EmployeesView, self).configure_form(f)
        employee = f.model_instance

        f.set_renderer('person', self.render_person)

        f.set_renderer('stores', self.render_stores)
        f.set_label('stores', "Stores") # TODO: should not be necessary
        if self.creating or self.editing:
            stores = self.get_possible_stores().all()
            store_values = [(s.uuid, six.text_type(s)) for s in stores]
            f.set_node('stores', colander.SchemaNode(colander.Set()))
            f.set_widget('stores', dfwidget.SelectWidget(multiple=True,
                                                         size=len(stores),
                                                         values=store_values))
            if self.editing:
                f.set_default('stores', [s.uuid for s in employee.stores])

        f.set_renderer('departments', self.render_departments)
        f.set_label('departments', "Departments") # TODO: should not be necessary
        if self.creating or self.editing:
            departments = self.get_possible_departments().all()
            dept_values = [(d.uuid, six.text_type(d)) for d in departments]
            f.set_node('departments', colander.SchemaNode(colander.Set()))
            f.set_widget('departments', dfwidget.SelectWidget(multiple=True,
                                                              size=len(departments),
                                                              values=dept_values))
            if self.editing:
                f.set_default('departments', [d.uuid for d in employee.departments])

        f.set_enum('status', self.enum.EMPLOYEE_STATUS)

        f.set_type('full_time_start', 'date_jquery')
        if self.editing:
            # TODO: this should not be needed (association proxy)
            f.set_default('full_time_start', employee.full_time_start)

        f.set_readonly('person')
        f.set_readonly('phone')
        f.set_readonly('email')

        f.set_label('display_name', "Short Name")
        f.set_label('phone', "Phone Number")
        f.set_label('email', "Email Address")

        if not self.viewing:
            f.remove_fields('first_name', 'last_name')

    def objectify(self, form, data=None):
        if data is None:
            data = form.validated
        employee = super(EmployeesView, self).objectify(form, data)
        self.update_stores(employee, data)
        self.update_departments(employee, data)
        return employee

    def update_stores(self, employee, data):
        old_stores = set([s.uuid for s in employee.stores])
        new_stores = data['stores']
        for uuid in new_stores:
            if uuid not in old_stores:
                employee._stores.append(model.EmployeeStore(store_uuid=uuid))
        for uuid in old_stores:
            if uuid not in new_stores:
                store = self.Session.query(model.Store).get(uuid)
                employee.stores.remove(store)

    def update_departments(self, employee, data):
        old_depts = set([d.uuid for d in employee.departments])
        new_depts = data['departments']
        for uuid in new_depts:
            if uuid not in old_depts:
                employee._departments.append(model.EmployeeDepartment(department_uuid=uuid))
        for uuid in old_depts:
            if uuid not in new_depts:
                dept = self.Session.query(model.Department).get(uuid)
                employee.departments.remove(dept)

    def get_possible_stores(self):
        return self.Session.query(model.Store)\
                           .order_by(model.Store.name)

    def get_possible_departments(self):
        return self.Session.query(model.Department)\
                           .order_by(model.Department.name)

    def render_person(self, employee, field):
        person = employee.person if employee else None
        if not person:
            return ""
        text = six.text_type(person)
        url = self.request.route_url('people.view', uuid=person.uuid)
        return tags.link_to(text, url)

    def render_stores(self, employee, field):
        stores = employee.stores if employee else None
        if not stores:
            return ""
        items = []
        for store in sorted(stores, key=six.text_type):
            items.append(HTML.tag('li', c=six.text_type(store)))
        return HTML.tag('ul', c=items)

    def render_departments(self, employee, field):
        departments = employee.departments if employee else None
        if not departments:
            return ""
        items = []
        for department in sorted(departments, key=six.text_type):
            items.append(HTML.tag('li', c=six.text_type(department)))
        return HTML.tag('ul', c=items)

    def get_version_child_classes(self):
        return [
            (model.Person, 'uuid', 'person_uuid'),
            (model.EmployeePhoneNumber, 'parent_uuid'),
            (model.EmployeeEmailAddress, 'parent_uuid'),
            (model.EmployeeStore, 'employee_uuid'),
            (model.EmployeeDepartment, 'employee_uuid'),
        ]


class EmployeesAutocomplete(AutocompleteView):
    """
    Autocomplete view for the Employee model, but restricted to return only
    results for current employees.
    """
    mapped_class = model.Person
    fieldname = 'display_name'

    def filter_query(self, q):
        return q.join(model.Employee)\
            .filter(model.Employee.status == self.enum.EMPLOYEE_STATUS_CURRENT)

    def value(self, person):
        return person.employee.uuid


def includeme(config):

    # autocomplete
    config.add_route('employees.autocomplete',  '/employees/autocomplete')
    config.add_view(EmployeesAutocomplete, route_name='employees.autocomplete',
                    renderer='json', permission='employees.list')

    EmployeesView.defaults(config)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2019 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Store Views
"""

from __future__ import unicode_literals, absolute_import

import sqlalchemy as sa

from rattail.db import model

import colander

from tailbone import grids
from tailbone.views import MasterView


class StoresView(MasterView):
    """
    Master view for the Store class.
    """
    model_class = model.Store
    has_versions = True

    grid_columns = [
        'id',
        'name',
        'phone',
        'email',
    ]

    labels = {
        'id':           "ID",
        'phone':        "Phone Number",
        'email':        "Email Address",
    }

    def configure_grid(self, g):
        super(StoresView, self).configure_grid(g)

        g.set_joiner('email', lambda q: q.outerjoin(model.StoreEmailAddress, sa.and_(
            model.StoreEmailAddress.parent_uuid == model.Store.uuid,
            model.StoreEmailAddress.preference == 1)))
        g.set_joiner('phone', lambda q: q.outerjoin(model.StorePhoneNumber, sa.and_(
            model.StorePhoneNumber.parent_uuid == model.Store.uuid,
            model.StorePhoneNumber.preference == 1)))

        g.set_filter('phone', model.StorePhoneNumber.number,
                     factory=grids.filters.AlchemyPhoneNumberFilter)
        g.filters['email'] = g.make_filter('email', model.StoreEmailAddress.address)
        g.filters['name'].default_active = True
        g.filters['name'].default_verb = 'contains'

        g.set_sorter('phone', model.StorePhoneNumber.number)
        g.set_sorter('email', model.StoreEmailAddress.address)
        g.set_sort_defaults('id')

        g.set_link('id')
        g.set_link('name')

    def configure_form(self, f):
        super(StoresView, self).configure_form(f)

        f.remove_field('employees')
        f.remove_field('phones')
        f.remove_field('emails')

        if self.creating:
            f.remove_field('phone')
            f.remove_field('email')
        else:
            f.set_readonly('phone')
            f.set_readonly('email')

    def get_version_child_classes(self):
        return [
            (model.StorePhoneNumber, 'parent_uuid'),
            (model.StoreEmailAddress, 'parent_uuid'),
        ]


def includeme(config):
    StoresView.defaults(config)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2019 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Views for 'ordering' (purchasing) batches
"""

from __future__ import unicode_literals, absolute_import

import os

import six
import openpyxl
from sqlalchemy import orm

from rattail.db import model, api
from rattail.core import Object
from rattail.time import localtime

from webhelpers2.html import tags

from tailbone.views.purchasing import PurchasingBatchView


class OrderingBatchView(PurchasingBatchView):
    """
    Master view for purchase order batches.
    """
    route_prefix = 'ordering'
    url_prefix = '/ordering'
    model_title = "Ordering Batch"
    model_title_plural = "Ordering Batches"
    index_title = "Ordering"
    mobile_creatable = True
    rows_editable = True
    mobile_rows_creatable = True
    mobile_rows_quickable = True
    mobile_rows_editable = True
    mobile_rows_deletable = True
    has_worksheet = True

    mobile_form_fields = [
        'vendor',
        'department',
        'date_ordered',
        'po_number',
        'po_total',
        'created',
        'created_by',
        'notes',
        'status_code',
        'complete',
        'executed',
        'executed_by',
    ]

    row_grid_columns = [
        'sequence',
        'upc',
        # 'item_id',
        'brand_name',
        'description',
        'size',
        'cases_ordered',
        'units_ordered',
        # 'cases_received',
        # 'units_received',
        'po_total',
        # 'invoice_total',
        # 'credits',
        'status_code',
    ]

    order_form_header_columns = [
        "UPC",
        "Brand",
        "Description",
        "Case",
        "Vend. Code",
        "Pref.",
        "Unit Cost",
    ]

    @property
    def batch_mode(self):
        return self.enum.PURCHASE_BATCH_MODE_ORDERING

    def configure_form(self, f):
        super(OrderingBatchView, self).configure_form(f)

        # purchase
        f.remove_field('purchase')

    def get_batch_kwargs(self, batch, mobile=False):
        kwargs = super(OrderingBatchView, self).get_batch_kwargs(batch, mobile=mobile)
        kwargs['ship_method'] = batch.ship_method
        kwargs['notes_to_vendor'] = batch.notes_to_vendor
        return kwargs

    def worksheet(self):
        """
        View for editing batch row data as an order form worksheet.
        """
        batch = self.get_instance()
        if batch.executed:
            return self.redirect(self.get_action_url('view', batch))

        # organize existing batch rows by product
        order_items = {}
        for row in batch.active_rows():
            order_items[row.product_uuid] = row

        # organize vendor catalog costs by dept / subdept
        departments = {}
        costs = self.get_order_form_costs(batch.vendor)
        costs = self.sort_order_form_costs(costs)
        for cost in costs:

            department = cost.product.department
            if department:
                departments.setdefault(department.uuid, department)
            else:
                if None not in departments:
                    department = Object(name='', number=None)
                    departments[None] = department
                department = departments[None]
            
            subdepartments = getattr(department, '_order_subdepartments', None)
            if subdepartments is None:
                subdepartments = department._order_subdepartments = {}

            subdepartment = cost.product.subdepartment
            if subdepartment:
                subdepartments.setdefault(subdepartment.uuid, subdepartment)
            else:
                if None not in subdepartments:
                    subdepartment = Object(name=None, number=None)
                    subdepartments[None] = subdepartment
                subdepartment = subdepartments[None]

            subdept_costs = getattr(subdepartment, '_order_costs', None)
            if subdept_costs is None:
                subdept_costs = subdepartment._order_costs = []
            subdept_costs.append(cost)
            cost._batchrow = order_items.get(cost.product_uuid)

            # do anything else needed to satisfy template display requirements etc.
            self.decorate_order_form_cost(cost)

        # fetch recent purchase history, sort/pad for template convenience
        history = self.get_order_form_history(batch, costs, 6)
        for i in range(6 - len(history)):
            history.append(None)
        history = list(reversed(history))

        title = self.get_instance_title(batch)
        order_date = batch.date_ordered
        if not order_date:
            order_date = localtime(self.rattail_config).date()
        return self.render_to_response('worksheet', {
            'batch': batch,
            'order_date': order_date,
            'instance': batch,
            'instance_title': title,
            'instance_url': self.get_action_url('view', batch),
            'vendor': batch.vendor,
            'departments': departments,
            'history': history,
            'get_upc': lambda p: p.upc.pretty() if p.upc else '',
            'header_columns': self.order_form_header_columns,
            'ignore_cases': not self.handler.allow_cases(),
        })

    def get_order_form_history(self, batch, costs, count):

        # fetch last 6 purchases for this vendor, organize line items by product
        history = []
        purchases = self.Session.query(model.Purchase)\
                                .filter(model.Purchase.vendor == batch.vendor)\
                                .filter(model.Purchase.status >= self.enum.PURCHASE_STATUS_ORDERED)\
                                .order_by(model.Purchase.date_ordered.desc(), model.Purchase.created.desc())\
                                .options(orm.joinedload(model.Purchase.items))
        for purchase in purchases[:count]:
            items = {}
            for item in purchase.items:
                items[item.product_uuid] = item
            history.append({'purchase': purchase, 'items': items})
        
        return history

    def get_order_form_costs(self, vendor):
        return self.Session.query(model.ProductCost)\
                           .join(model.Product)\
                           .outerjoin(model.Brand)\
                           .filter(model.ProductCost.vendor == vendor)\
                           .options(orm.joinedload(model.ProductCost.product)\
                                    .joinedload(model.Product.department))\
                           .options(orm.joinedload(model.ProductCost.product)\
                                    .joinedload(model.Product.subdepartment))

    def sort_order_form_costs(self, costs):
        return costs.order_by(model.Brand.name,
                              model.Product.description,
                              model.Product.size)

    def decorate_order_form_cost(self, cost):
        pass

    def worksheet_update(self):
        """
        Handles AJAX requests to update current batch, from Order Form view.
        """
        batch = self.get_instance()

        cases_ordered = self.request.POST.get('cases_ordered', '0')
        if not cases_ordered or not cases_ordered.isdigit():
            return {'error': "Invalid value for cases ordered: {}".format(cases_ordered)}
        cases_ordered = int(cases_ordered)
        if cases_ordered >= 100000: # TODO: really this depends on underlying column
            return {'error': "Invalid value for cases ordered: {}".format(cases_ordered)}

        units_ordered = self.request.POST.get('units_ordered', '0')
        if not units_ordered or not units_ordered.isdigit():
            return {'error': "Invalid value for units ordered: {}".format(units_ordered)}
        units_ordered = int(units_ordered)
        if units_ordered >= 100000: # TODO: really this depends on underlying column
            return {'error': "Invalid value for units ordered: {}".format(units_ordered)}

        uuid = self.request.POST.get('product_uuid')
        product = self.Session.query(model.Product).get(uuid) if uuid else None
        if not product:
            return {'error': "Product not found"}

        row = None
        rows = [r for r in batch.data_rows if r.product_uuid == uuid]
        if rows:
            assert len(rows) == 1
            row = rows[0]
            if row.po_total and not row.removed:
                batch.po_total -= row.po_total
            if cases_ordered or units_ordered:
                row.cases_ordered = cases_ordered or None
                row.units_ordered = units_ordered or None
                if row.removed:
                    row.removed = False
                    batch.rowcount += 1
                self.handler.refresh_row(row)
                if row.po_unit_cost:
                    row.po_total = row.po_unit_cost * self.handler.get_units_ordered(row)
                    batch.po_total = (batch.po_total or 0) + row.po_total
            else:
                row.removed = True

        elif cases_ordered or units_ordered:
            row = model.PurchaseBatchRow()
            row.product = product
            row.cases_ordered = cases_ordered or None
            row.units_ordered = units_ordered or None
            self.handler.add_row(batch, row)
            if row.po_unit_cost:
                row.po_total = row.po_unit_cost * self.handler.get_units_ordered(row)
                batch.po_total = (batch.po_total or 0) + row.po_total

        return {
            'row_cases_ordered': '' if not row or row.removed else int(row.cases_ordered or 0),
            'row_units_ordered': '' if not row or row.removed else int(row.units_ordered or 0),
            'row_po_total': '' if not row or row.removed else '${:0,.2f}'.format(row.po_total or 0),
            'batch_po_total': '${:0,.2f}'.format(batch.po_total or 0),
        }

    def render_mobile_listitem(self, batch, i):
        return "({}) {} on {} for ${:0,.2f}".format(batch.id_str, batch.vendor,
                                                    batch.date_ordered, batch.po_total or 0)

    def mobile_create(self):
        """
        Mobile view for creating a new ordering batch
        """
        mode = self.batch_mode
        data = {'mode': mode}

        vendor = None
        if self.request.method == 'POST' and self.request.POST.get('vendor'):
            vendor = self.Session.query(model.Vendor).get(self.request.POST['vendor'])
            if vendor:

                # fetch first to avoid flush below
                store = self.rattail_config.get_store(self.Session())

                batch = self.model_class()
                batch.mode = mode
                batch.vendor = vendor
                batch.store = store
                batch.buyer = self.request.user.employee
                batch.created_by = self.request.user
                batch.po_total = 0
                kwargs = self.get_batch_kwargs(batch, mobile=True)
                batch = self.handler.make_batch(self.Session(), **kwargs)
                if self.handler.should_populate(batch):
                    self.handler.populate(batch)
                return self.redirect(self.request.route_url('mobile.ordering.view', uuid=batch.uuid))

        data['index_title'] = self.get_index_title()
        data['index_url'] = self.get_index_url(mobile=True)
        data['mode_title'] = self.enum.PURCHASE_BATCH_MODE[mode].capitalize()

        data['vendor_use_autocomplete'] = self.rattail_config.getbool(
            'rattail', 'vendor.use_autocomplete', default=True)
        if not data['vendor_use_autocomplete']:
            vendors = self.Session.query(model.Vendor)\
                                  .order_by(model.Vendor.name)
            options = [(tags.Option(vendor.name, vendor.uuid))
                       for vendor in vendors]
            options.insert(0, tags.Option("(please choose)", ''))
            data['vendor_options'] = options

        return self.render_to_response('create', data, mobile=True)

    def configure_mobile_row_form(self, f):
        super(OrderingBatchView, self).configure_mobile_row_form(f)
        if self.editing:
            # TODO: probably should take `allow_cases` into account here...
            f.focus_spec = '[name="units_ordered"]'

    def download_excel(self):
        """
        Download ordering batch as Excel spreadsheet.
        """
        batch = self.get_instance()

        # populate Excel worksheet
        workbook = openpyxl.Workbook()
        worksheet = workbook.active
        worksheet.title = "Purchase Order"
        worksheet.append(["Store", "Vendor", "Date ordered"])
        worksheet.append([batch.store.name, batch.vendor.name, batch.date_ordered.strftime('%m/%d/%Y')])
        worksheet.append([])
        worksheet.append(['vendor_code', 'upc', 'brand_name', 'description', 'cases_ordered', 'units_ordered'])
        for row in batch.active_rows():
            worksheet.append([row.vendor_code, six.text_type(row.upc), row.brand_name,
                              '{} {}'.format(row.description, row.size),
                              row.cases_ordered, row.units_ordered])

        # write Excel file to batch data dir
        filedir = batch.filedir(self.rattail_config)
        if not os.path.exists(filedir):
            os.makedirs(filedir)
        filename = 'PO.{}.xlsx'.format(batch.id_str)
        path = batch.filepath(self.rattail_config, filename)
        workbook.save(path)

        return self.file_response(path)

    @classmethod
    def _ordering_defaults(cls, config):
        route_prefix = cls.get_route_prefix()
        url_prefix = cls.get_url_prefix()
        permission_prefix = cls.get_permission_prefix()
        model_title = cls.get_model_title()
        model_title_plural = cls.get_model_title_plural()

        # fix permission group label
        config.add_tailbone_permission_group(permission_prefix, model_title_plural)

        # download as Excel
        config.add_route('{}.download_excel'.format(route_prefix), '{}/{{uuid}}/excel'.format(url_prefix))
        config.add_view(cls, attr='download_excel', route_name='{}.download_excel'.format(route_prefix),
                        permission='{}.download_excel'.format(permission_prefix))
        config.add_tailbone_permission(permission_prefix, '{}.download_excel'.format(permission_prefix),
                                       "Download {} as Excel".format(model_title))

    @classmethod
    def defaults(cls, config):
        cls._ordering_defaults(config)
        cls._purchasing_defaults(config)
        cls._batch_defaults(config)
        cls._defaults(config)


def includeme(config):
    OrderingBatchView.defaults(config)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2018 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Tailbone Web API - Auth Views
"""

from __future__ import unicode_literals, absolute_import

from rattail.db.auth import authenticate_user

from tailbone.api import APIView, api
from tailbone.db import Session
from tailbone.auth import login_user, logout_user


class AuthenticationView(APIView):

    def user_info(self, user):
        return {
            'ok': True,
            'user': {
                'uuid': user.uuid,
                'username': user.username,
                'display_name': user.display_name,
            },
        }

    @api
    def check_session(self):
        """
        View to serve as "no-op" / ping action to check current user's session.
        This will establish a server-side web session for the user if none
        exists.  Note that this also resets the user's session timer.
        """
        if self.request.user:
            return self.user_info(self.request.user)
        return {}

    @api
    def login(self):
        """
        API login view.
        """
        if self.request.method == 'OPTIONS':
            return self.request.response

        username = self.request.json.get('username')
        password = self.request.json.get('password')
        if not (username and password):
            return {'error': "Invalid username or password"}

        user = authenticate_user(Session(), username, password)
        if not user:
            return {'error': "Invalid username or password"}

        login_user(self.request, user)
        return self.user_info(user)

    @api
    def logout(self):
        """
        API logout view.
        """
        if self.request.method == 'OPTIONS':
            return self.request.response

        logout_user(self.request)
        return {'ok': True}

    @classmethod
    def defaults(cls, config):

        # session
        config.add_route('api.session', '/session', request_method='GET')
        config.add_view(cls, attr='check_session', route_name='api.session', renderer='json')

        # login
        config.add_route('api.login', '/login', request_method=('OPTIONS', 'POST'))
        config.add_view(cls, attr='login', route_name='api.login', renderer='json')

        # logout
        config.add_route('api.logout', '/logout', request_method=('OPTIONS', 'POST'))
        config.add_view(cls, attr='logout', route_name='api.logout', renderer='json')


def includeme(config):
    AuthenticationView.defaults(config)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2018 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Views for pricing batches
"""

from __future__ import unicode_literals, absolute_import

from rattail.db import model

from webhelpers2.html import tags

from tailbone.views.batch import BatchMasterView


class PricingBatchView(BatchMasterView):
    """
    Master view for pricing batches.
    """
    model_class = model.PricingBatch
    model_row_class = model.PricingBatchRow
    default_handler_spec = 'rattail.batch.pricing:PricingBatchHandler'
    model_title_plural = "Pricing Batches"
    route_prefix = 'batch.pricing'
    url_prefix = '/batches/pricing'
    template_prefix = '/batch/pricing'
    creatable = True
    downloadable = True
    bulk_deletable = True
    rows_editable = True
    rows_bulk_deletable = True

    labels = {
        'min_diff_threshold': "Min $ Diff",
        'min_diff_percent': "Min % Diff",
    }

    grid_columns = [
        'id',
        'description',
        'created',
        'created_by',
        'rowcount',
        # 'status_code',
        # 'complete',
        'executed',
        'executed_by',
    ]

    form_fields = [
        'id',
        'input_filename',
        'description',
        'min_diff_threshold',
        'min_diff_percent',
        'calculate_for_manual',
        'notes',
        'created',
        'created_by',
        'rowcount',
        'executed',
        'executed_by',
    ]

    row_labels = {
        'upc': "UPC",
        'vendor_id': "Vendor ID",
        'regular_unit_cost': "Reg. Cost",
        'price_diff': "$ Diff",
        'price_diff_percent': "% Diff",
        'brand_name': "Brand",
        'price_markup': "Markup",
        'manually_priced': "Manual",
    }

    row_grid_columns = [
        'sequence',
        'upc',
        'brand_name',
        'description',
        'size',
        'vendor_id',
        'discounted_unit_cost',
        'old_price',
        'new_price',
        'price_margin',
        'price_diff',
        'price_diff_percent',
        'manually_priced',
        'status_code',
    ]

    row_form_fields = [
        'sequence',
        'product',
        'upc',
        'brand_name',
        'description',
        'size',
        'department_number',
        'department_name',
        'subdepartment_number',
        'subdepartment_name',
        'family_code',
        'report_code',
        'alternate_code',
        'vendor',
        'vendor_item_code',
        'regular_unit_cost',
        'discounted_unit_cost',
        'suggested_price',
        'old_price',
        'new_price',
        'price_diff',
        'price_diff_percent',
        'price_markup',
        'price_margin',
        'old_price_margin',
        'margin_diff',
        'status_code',
        'status_text',
    ]

    def configure_form(self, f):
        super(PricingBatchView, self).configure_form(f)

        f.set_type('min_diff_threshold', 'currency')

        # input_filename
        if self.creating:
            f.set_type('input_filename', 'file')
        else:
            f.set_readonly('input_filename')
            f.set_renderer('input_filename', self.render_downloadable_file)

    def get_batch_kwargs(self, batch, mobile=False):
        kwargs = super(PricingBatchView, self).get_batch_kwargs(batch, mobile=mobile)
        kwargs['min_diff_threshold'] = batch.min_diff_threshold
        kwargs['min_diff_percent'] = batch.min_diff_percent
        kwargs['calculate_for_manual'] = batch.calculate_for_manual
        return kwargs

    def configure_row_grid(self, g):
        super(PricingBatchView, self).configure_row_grid(g)

        g.set_joiner('vendor_id', lambda q: q.outerjoin(model.Vendor))
        g.set_sorter('vendor_id', model.Vendor.id)
        g.set_filter('vendor_id', model.Vendor.id)
        g.set_renderer('vendor_id', self.render_vendor_id)

        g.set_type('old_price', 'currency')
        g.set_type('new_price', 'currency')
        g.set_type('price_diff', 'currency')

    def render_vendor_id(self, row, field):
        vendor_id = row.vendor.id if row.vendor else None
        if not vendor_id:
            return ""
        return vendor_id

    def row_grid_extra_class(self, row, i):
        if row.status_code == row.STATUS_CANNOT_CALCULATE_PRICE:
            return 'warning'
        if row.status_code in (row.STATUS_PRICE_INCREASE, row.STATUS_PRICE_DECREASE):
            return 'notice'

    def configure_row_form(self, f):
        super(PricingBatchView, self).configure_row_form(f)

        # readonly fields
        f.set_readonly('product')
        f.set_readonly('upc')
        f.set_readonly('brand_name')
        f.set_readonly('description')
        f.set_readonly('size')
        f.set_readonly('department_number')
        f.set_readonly('department_name')
        f.set_readonly('vendor')

        # product
        f.set_renderer('product', self.render_product)

        # currency fields
        f.set_type('suggested_price', 'currency')
        f.set_type('old_price', 'currency')
        f.set_type('new_price', 'currency')
        f.set_type('price_diff', 'currency')

        # vendor
        f.set_renderer('vendor', self.render_vendor)

    def render_vendor(self, row, field):
        vendor = row.vendor
        if not vendor:
            return ""
        text = "({}) {}".format(vendor.id, vendor.name)
        url = self.request.route_url('vendors.view', uuid=vendor.uuid)
        return tags.link_to(text, url)

    def get_row_csv_fields(self):
        fields = super(PricingBatchView, self).get_row_csv_fields()

        if 'vendor_uuid' in fields:
            i = fields.index('vendor_uuid')
            fields.insert(i + 1, 'vendor_id')
            fields.insert(i + 2, 'vendor_abbreviation')
            fields.insert(i + 3, 'vendor_name')
        else:
            fields.append('vendor_id')
            fields.append('vendor_abbreviation')
            fields.append('vendor_name')

        return fields

    def get_row_csv_row(self, row, fields):
        csvrow = super(PricingBatchView, self).get_row_csv_row(row, fields)

        vendor = row.vendor
        if 'vendor_id' in fields:
            csvrow['vendor_id'] = (vendor.id or '') if vendor else ''
        if 'vendor_abbreviation' in fields:
            csvrow['vendor_abbreviation'] = (vendor.abbreviation or '') if vendor else ''
        if 'vendor_name' in fields:
            csvrow['vendor_name'] = (vendor.name or '') if vendor else ''

        return csvrow


def includeme(config):
    PricingBatchView.defaults(config)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2018 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Views for employee schedules
"""

from __future__ import unicode_literals, absolute_import

import datetime

from rattail.db import model
from rattail.time import localtime, make_utc, get_sunday

from tailbone.db import Session
from tailbone.views.shifts.lib import TimeSheetView


class ScheduleView(TimeSheetView):
    """
    Simple view for current user's schedule.
    """
    key = 'schedule'
    model_class = model.ScheduledShift

    def edit(self):
        """
        View for editing (full) schedule.
        """
        # first check if we should clear the schedule
        if self.request.method == 'POST' and self.request.POST.get('clear-schedule') == 'clear':
            count = self.clear_schedule()
            self.request.session.flash("Removed {} shifts from current schedule.".format(count))
            return self.redirect(self.request.route_url('schedule.edit'))

        # okay then, check if we should copy data from another week
        if self.request.method == 'POST' and self.request.POST.get('copy-week'):
            sunday, copied = self.copy_schedule()
            self.request.session.flash("Copied {} shifts from week of {}".format(copied, sunday.strftime('%m/%d/%Y')))
            return self.redirect(self.request.route_url('schedule.edit'))

        # okay then, process filters; redirect if any were received
        context = self.get_timesheet_context()
        form = self.make_full_filter_form(context)
        self.process_filter_form(form)

        # okay then, maybe process saved shift data
        if self.request.method == 'POST':

            # organize form data by uuid / field
            fields = ['employee_uuid', 'store_uuid', 'start_time', 'end_time', 'delete']
            data = dict([(f, {}) for f in fields])
            for key in self.request.POST:
                for field in fields:
                    if key.startswith('{}-'.format(field)):
                        uuid = key[len('{}-'.format(field)):]
                        if uuid:
                            data[field][uuid] = self.request.POST[key]
                        break

            # apply delete operations
            deleted = []
            for uuid, value in data['delete'].items():
                if value == 'delete':
                    shift = Session.query(model.ScheduledShift).get(uuid)
                    if shift:
                        Session.delete(shift)
                        deleted.append(uuid)

            # apply create / update operations
            created = {}
            updated = {}
            time_format = '%a %d %b %Y %I:%M %p'
            for uuid, employee_uuid in data['start_time'].items():
                if uuid in deleted:
                    continue
                if uuid.startswith('new-'):
                    shift = model.ScheduledShift()
                    shift.employee_uuid = data['employee_uuid'][uuid]
                    if 'store_uuid' in data and uuid in data['store_uuid']:
                        shift.store_uuid = data['store_uuid'][uuid]
                    else:
                        shift.store_uuid = context['store'].uuid if context['store'] else None
                    Session.add(shift)
                    created[uuid] = shift
                else:
                    shift = Session.query(model.ScheduledShift).get(uuid)
                    assert shift
                    updated[uuid] = shift
                start_time = datetime.datetime.strptime(data['start_time'][uuid], time_format)
                shift.start_time = make_utc(localtime(self.rattail_config, start_time))
                end_time = datetime.datetime.strptime(data['end_time'][uuid], time_format)
                shift.end_time = make_utc(localtime(self.rattail_config, end_time))

            self.request.session.flash("Changes were applied: created {}, updated {}, "
                                       "deleted {} Scheduled Shifts".format(
                                           len(created), len(updated), len(deleted)))
            return self.redirect(self.request.route_url('schedule.edit'))

        context['form'] = form
        context['page_title'] = "Edit Schedule"
        context['allow_clear'] = self.rattail_config.getbool('tailbone', 'schedule.allow_clear',
                                                             default=True)
        return self.render_full(**context)

    def clear_schedule(self):
        deleted = 0
        context = self.get_timesheet_context()
        if context['employees']:
            sunday = datetime.datetime.combine(context['date'], datetime.time(0))
            start_time = localtime(self.rattail_config, sunday)
            end_time = localtime(self.rattail_config, sunday + datetime.timedelta(days=7))
            shifts = Session.query(model.ScheduledShift)\
                            .filter(model.ScheduledShift.employee_uuid.in_([e.uuid for e in context['employees']]))\
                            .filter(model.ScheduledShift.start_time >= make_utc(start_time))\
                            .filter(model.ScheduledShift.end_time < make_utc(end_time))
            for shift in shifts:
                Session.delete(shift)
                deleted += 1
        return deleted

    def copy_schedule(self):
        """
        Clear current schedule, then copy shift data from another week.
        """
        try:
            sunday = datetime.datetime.strptime(self.request.POST['copy-week'], '%m/%d/%Y').date()
        except ValueError as error:
            self.request.session.flash("Invalid date specified: {}: {}".format(type(error), error), 'error')
            raise self.redirect(self.request.route_url('schedule.edit'))
        sunday = get_sunday(sunday)
        context = self.get_timesheet_context()
        if sunday == context['date']:
            self.request.session.flash("Cannot copy schedule from same week; please specify a different week.", 'error')
            raise self.redirect(self.request.route_url('schedule.edit'))

        self.clear_schedule()

        copied = 0
        if context['employees']:
            offset = context['date'] - sunday
            sunday = datetime.datetime.combine(sunday, datetime.time(0))
            start_time = localtime(self.rattail_config, sunday)
            end_time = localtime(self.rattail_config, sunday + datetime.timedelta(days=7))
            shifts = Session.query(model.ScheduledShift)\
                            .filter(model.ScheduledShift.employee_uuid.in_([e.uuid for e in context['employees']]))\
                            .filter(model.ScheduledShift.start_time >= make_utc(start_time))\
                            .filter(model.ScheduledShift.end_time < make_utc(end_time))
            for shift in shifts:

                # must calculate new times using date as base, b/c of daylight savings
                start_time = localtime(self.rattail_config, shift.start_time, from_utc=True)
                start_time = datetime.datetime.combine(start_time.date() + offset, start_time.time())
                start_time = localtime(self.rattail_config, start_time)
                end_time = localtime(self.rattail_config, shift.end_time, from_utc=True)
                end_time = datetime.datetime.combine(end_time.date() + offset, end_time.time())
                end_time = localtime(self.rattail_config, end_time)

                Session.add(model.ScheduledShift(
                    employee_uuid=shift.employee_uuid,
                    store_uuid=shift.store_uuid,
                    start_time=make_utc(start_time),
                    end_time=make_utc(end_time),
                ))
                copied += 1

        return sunday, copied

    @classmethod
    def defaults(cls, config):
        cls._defaults(config)

        # edit schedule
        config.add_route('schedule.edit', '/schedule/edit')
        config.add_view(cls, attr='edit', route_name='schedule.edit',
                        renderer='/shifts/schedule_edit.mako',
                        permission='schedule.edit')
        config.add_tailbone_permission('schedule', 'schedule.edit', "Edit full schedule")

        # printing "any" schedule requires this permission
        config.add_tailbone_permission('schedule', 'schedule.print', "Print schedule")

        # print full schedule
        config.add_route('schedule.print', '/schedule/print')
        config.add_view(cls, attr='full', route_name='schedule.print',
                        renderer='/shifts/schedule_print.mako',
                        permission='schedule.print')

        # print employee schedule
        config.add_route('schedule.employee.print', '/schedule/employee/print')
        config.add_view(cls, attr='employee', route_name='schedule.employee.print',
                        renderer='/shifts/schedule_print_employee.mako',
                        permission='schedule.print')


def includeme(config):
    ScheduleView.defaults(config)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2018 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Customer Order Views
"""

from __future__ import unicode_literals, absolute_import

import six
from sqlalchemy import orm

from rattail.db import model

from webhelpers2.html import tags

from tailbone.db import Session
from tailbone.views import MasterView


class CustomerOrdersView(MasterView):
    """
    Master view for customer orders
    """
    model_class = model.CustomerOrder
    route_prefix = 'custorders'
    creatable = False
    editable = False
    deletable = False

    grid_columns = [
        'id',
        'customer',
        'person',
        'created',
        'status_code',
    ]

    form_fields = [
        'id',
        'customer',
        'person',
        'created',
        'status_code',
    ]

    def query(self, session):
        return session.query(model.CustomerOrder)\
                      .options(orm.joinedload(model.CustomerOrder.customer))

    def configure_grid(self, g):
        super(CustomerOrdersView, self).configure_grid(g)

        g.set_joiner('customer', lambda q: q.outerjoin(model.Customer))
        g.set_joiner('person', lambda q: q.outerjoin(model.Person))

        g.filters['customer'] = g.make_filter('customer', model.Customer.name,
                                              label="Customer Name",
                                              default_active=True,
                                              default_verb='contains')
        g.filters['person'] = g.make_filter('person', model.Person.display_name,
                                            label="Person Name",
                                            default_active=True,
                                            default_verb='contains')

        g.set_sorter('customer', model.Customer.name)
        g.set_sorter('person', model.Person.display_name)

        g.set_sort_defaults('created', 'desc')

        # TODO: enum choices renderer
        g.set_label('status_code', "Status")
        g.set_label('id', "ID")

    def configure_form(self, f):
        super(CustomerOrdersView, self).configure_form(f)

        # id
        f.set_readonly('id')
        f.set_label('id', "ID")

        # person
        f.set_renderer('person', self.render_person)

        # created
        f.set_readonly('created')

        # label overrides
        f.set_label('status_code', "Status")

    def render_person(self, order, field):
        person = order.person
        if not person:
            return ""
        text = six.text_type(person)
        url = self.request.route_url('people.view', uuid=person.uuid)
        return tags.link_to(text, url)


def includeme(config):
    CustomerOrdersView.defaults(config)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2019 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Views for tempmon clients
"""

from __future__ import unicode_literals, absolute_import

import subprocess

from rattail.config import parse_list
from rattail_tempmon.db import model as tempmon

import colander
from webhelpers2.html import HTML, tags

from tailbone import forms
from tailbone.views.tempmon import MasterView
from tailbone.util import raw_datetime


class TempmonClientView(MasterView):
    """
    Master view for tempmon clients.
    """
    model_class = tempmon.Client
    model_title = "TempMon Client"
    model_title_plural = "TempMon Clients"
    route_prefix = 'tempmon.clients'
    url_prefix = '/tempmon/clients'

    has_rows = True
    model_row_class = tempmon.Reading

    grid_columns = [
        'config_key',
        'hostname',
        'location',
        'delay',
        'enabled',
        'online',
        'archived',
    ]

    form_fields = [
        'config_key',
        'hostname',
        'location',
        'disk_type',
        'delay',
        'appliances',
        'probes',
        'notes',
        'enabled',
        'online',
        'archived',
    ]

    row_grid_columns = [
        'probe',
        'degrees_f',
        'taken',
    ]

    def configure_grid(self, g):
        super(TempmonClientView, self).configure_grid(g)

        # config_key
        g.set_label('config_key', "Key")
        g.set_sort_defaults('config_key')
        g.set_link('config_key')

        # hostname
        g.filters['hostname'].default_active = True
        g.filters['hostname'].default_verb = 'contains'
        g.set_link('hostname')

        # location
        g.filters['location'].default_active = True
        g.filters['location'].default_verb = 'contains'
        g.set_link('location')

        # disk_type
        g.set_enum('disk_type', self.enum.TEMPMON_DISK_TYPE)

        # enabled
        g.set_renderer('enabled', self.render_enabled_grid)

        # archived
        g.filters['archived'].default_active = True
        g.filters['archived'].default_verb = 'is_false'

    def render_enabled_grid(self, client, field):
        if client.enabled:
            return "Yes"
        return "No"

    def configure_form(self, f):
        super(TempmonClientView, self).configure_form(f)

        # config_key
        f.set_validator('config_key', self.unique_config_key)

        # disk_type
        f.set_enum('disk_type', self.enum.TEMPMON_DISK_TYPE)
        f.widgets['disk_type'].values.insert(0, ('', "(unknown)"))

        # delay
        f.set_helptext('delay', tempmon.Client.delay.__doc__)

        # appliances
        if self.viewing:
            f.set_renderer('appliances', self.render_appliances)
        else:
            f.remove_field('appliances')

        # probes
        if self.viewing:
            f.set_renderer('probes', self.render_probes)
        else:
            f.remove_field('probes')

        # notes
        f.set_type('notes', 'text')

        # enabled
        if self.creating or self.editing:
            f.set_node('enabled', forms.types.DateTimeBoolean())
        else:
            f.set_renderer('enabled', self.render_enabled_form)
        f.set_helptext('enabled', tempmon.Client.enabled.__doc__)

        # online
        if self.creating or self.editing:
            f.remove_field('online')
        else:
            f.set_helptext('online', tempmon.Client.online.__doc__)

        # archived
        f.set_helptext('archived', tempmon.Client.archived.__doc__)

    def objectify(self, form, data=None):

        # this is a hack to prevent updates to the 'enabled' timestamp, when
        # simple edits are being done to the client.  i.e. we do want to set
        # the timestamp when it was previously null, but not otherwise.
        if self.editing:
            data = dict(data or form.validated)
            if data['enabled'] and form.model_instance.enabled:
                data['enabled'] = form.model_instance.enabled

        return super(TempmonClientView, self).objectify(form, data=data)

    def unique_config_key(self, node, value):
        query = self.Session.query(tempmon.Client)\
                            .filter(tempmon.Client.config_key == value)
        if self.editing:
            client = self.get_instance()
            query = query.filter(tempmon.Client.uuid != client.uuid)
        if query.count():
            raise colander.Invalid(node, "Config key must be unique")

    def render_appliances(self, client, field):
        appliances = {}
        for probe in client.probes:
            if probe.appliance and probe.appliance.uuid not in appliances:
                appliances[probe.appliance.uuid] = probe.appliance

        if not appliances:
            return ""

        appliances = sorted(appliances.values(), key=lambda a: a.name)
        items = [HTML.tag('li', c=[tags.link_to(a.name, self.request.route_url('tempmon.appliances.view', uuid=a.uuid))])
                 for a in appliances]
        return HTML.tag('ul', c=items)

    def render_enabled_form(self, client, field):
        if client.enabled:
            return raw_datetime(self.rattail_config, client.enabled)
        return "No"

    def delete_instance(self, client):
        # bulk-delete all readings first
        readings = self.Session.query(tempmon.Reading)\
                               .filter(tempmon.Reading.client == client)
        readings.delete(synchronize_session=False)
        self.Session.flush()
        self.Session.refresh(client)

        # Flush immediately to force any pending integrity errors etc.; that
        # way we don't set flash message until we know we have success.
        self.Session.delete(client)
        self.Session.flush()

    def get_row_data(self, client):
        query = self.Session.query(tempmon.Reading)\
                            .join(tempmon.Probe)\
                            .filter(tempmon.Reading.client == client)
        return query

    def get_parent(self, reading):
        return reading.client

    def configure_row_grid(self, g):
        super(TempmonClientView, self).configure_row_grid(g)

        # probe
        g.set_filter('probe', tempmon.Probe.description)
        g.set_sorter('probe', tempmon.Probe.description)

        g.set_sort_defaults('taken', 'desc')

    def restartable_client(self, client):
        cmd = self.get_restart_cmd(client)
        return bool(cmd)

    def restart(self):
        client = self.get_instance()
        if self.restartable_client(client):
            try:
                subprocess.check_output(self.get_restart_cmd(client),
                                        stderr=subprocess.STDOUT)
            except subprocess.CalledProcessError as error:
                self.request.session.flash("Failed to restart client: {}".format(error.output), 'error')
            else:
                self.request.session.flash("Client has been restarted: {}".format(
                    self.get_instance_title(client)))
        else:
            self.request.session.flash("Restart not supported for client: {}".format(client), 'error')
        return self.redirect(self.get_action_url('view', client))

    def get_restart_cmd(self, client):
        name = 'rattail.tempmon.client.restart'
        cmd = self.rattail_config.get('rattail.tempmon', 'client.restart.{}'.format(client.config_key))
        if not cmd:
            cmd = self.rattail_config.get('rattail.tempmon', 'client.restart')
        if cmd:
            cmd = cmd.format(hostname=client.hostname)
            return parse_list(cmd)

    @classmethod
    def defaults(cls, config):
        route_prefix = cls.get_route_prefix()
        url_prefix = cls.get_url_prefix()
        permission_prefix = cls.get_permission_prefix()
        model_key = cls.get_model_key()
        model_title = cls.get_model_title()

        cls._defaults(config)

        # restart tempmon client
        config.add_tailbone_permission(permission_prefix, '{}.restart'.format(permission_prefix),
                                       "Restart a {}".format(model_title))
        config.add_route('{}.restart'.format(route_prefix), '{}/{{{}}}/restart'.format(url_prefix, model_key))
        config.add_view(cls, attr='restart', route_name='{}.restart'.format(route_prefix),
                        permission='{}.restart'.format(permission_prefix))


def includeme(config):
    TempmonClientView.defaults(config)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2019 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Reporting views
"""

from __future__ import unicode_literals, absolute_import

import re
import datetime
import logging

import six

import rattail
from rattail.db import model, Session as RattailSession
from rattail.files import resource_path
from rattail.time import localtime
from rattail.reporting import get_report_handler
from rattail.threads import Thread

import colander
from mako.template import Template
from pyramid.response import Response
from webhelpers2.html import HTML

from tailbone import forms, grids
from tailbone.db import Session
from tailbone.views import View
from tailbone.views.exports import ExportMasterView
from tailbone.progress import SessionProgress


plu_upc_pattern = re.compile(r'^000000000(\d{5})$')
weighted_upc_pattern = re.compile(r'^002(\d{5})00000\d$')

log = logging.getLogger(__name__)


def get_upc(product):
    """
    UPC formatter.  Strips PLUs to bare number, and adds "minus check digit"
    for non-PLU UPCs.
    """
    upc = six.text_type(product.upc)
    m = plu_upc_pattern.match(upc)
    if m:
        return six.text_type(int(m.group(1)))
    m = weighted_upc_pattern.match(upc)
    if m:
        return six.text_type(int(m.group(1)))
    return '{0}-{1}'.format(upc[:-1], upc[-1])


class OrderingWorksheet(View):
    """
    This is the "Ordering Worksheet" report.
    """

    report_template_path = 'tailbone:reports/ordering_worksheet.mako'

    upc_getter = staticmethod(get_upc)

    def __call__(self):
        if self.request.params.get('vendor'):
            vendor = Session.query(model.Vendor).get(self.request.params['vendor'])
            if vendor:
                departments = []
                uuids = self.request.params.get('departments')
                if uuids:
                    for uuid in uuids.split(','):
                        dept = Session.query(model.Department).get(uuid)
                        if dept:
                            departments.append(dept)
                preferred_only = self.request.params.get('preferred_only') == '1'
                body = self.write_report(vendor, departments, preferred_only)
                response = Response(content_type='text/html')
                response.headers['Content-Length'] = len(body)
                response.headers['Content-Disposition'] = 'attachment; filename=ordering.html'
                response.text = body
                return response
        return {}

    def write_report(self, vendor, departments, preferred_only):
        """
        Rendering engine for the ordering worksheet report.
        """

        q = Session.query(model.ProductCost)
        q = q.join(model.Product)
        q = q.filter(model.Product.deleted == False)
        q = q.filter(model.ProductCost.vendor == vendor)
        q = q.filter(model.Product.department_uuid.in_([x.uuid for x in departments]))
        if preferred_only:
            q = q.filter(model.ProductCost.preference == 1)

        costs = {}
        for cost in q:
            dept = cost.product.department
            subdept = cost.product.subdepartment
            costs.setdefault(dept, {})
            costs[dept].setdefault(subdept, [])
            costs[dept][subdept].append(cost)

        def cost_sort_key(cost):
            product = cost.product
            brand = product.brand.name if product.brand else ''
            key = '{0} {1}'.format(brand, product.description)
            return key

        now = localtime(self.request.rattail_config)
        data = dict(
            vendor=vendor,
            costs=costs,
            cost_sort_key=cost_sort_key,
            date=now.strftime('%a %d %b %Y'),
            time=now.strftime('%I:%M %p'),
            get_upc=self.upc_getter,
            rattail=rattail,
            )

        template_path = resource_path(self.report_template_path)
        template = Template(filename=template_path)
        return template.render(**data)


class InventoryWorksheet(View):
    """
    This is the "Inventory Worksheet" report.
    """

    report_template_path = 'tailbone:reports/inventory_worksheet.mako'

    upc_getter = staticmethod(get_upc)

    def __call__(self):
        """
        This is the "Inventory Worksheet" report.
        """

        departments = Session.query(model.Department)

        if self.request.params.get('department'):
            department = departments.get(self.request.params['department'])
            if department:
                body = self.write_report(department)
                response = Response(content_type=str('text/html'))
                response.headers[str('Content-Length')] = len(body)
                response.headers[str('Content-Disposition')] = str('attachment; filename=inventory.html')
                response.text = body
                return response

        departments = departments.order_by(model.Department.name)
        departments = departments.all()
        return{'departments': departments}

    def write_report(self, department):
        """
        Generates the Inventory Worksheet report.
        """

        def get_products(subdepartment):
            q = Session.query(model.Product)
            q = q.outerjoin(model.Brand)
            q = q.filter(model.Product.deleted == False)
            q = q.filter(model.Product.subdepartment == subdepartment)
            if self.request.params.get('weighted-only'):
                q = q.filter(model.Product.weighed == True)
            if self.request.params.get('exclude-not-for-sale'):
                q = q.filter(model.Product.not_for_sale == False)
            q = q.order_by(model.Brand.name, model.Product.description)
            return q.all()

        now = localtime(self.request.rattail_config)
        data = dict(
            date=now.strftime('%a %d %b %Y'),
            time=now.strftime('%I:%M %p'),
            department=department,
            get_products=get_products,
            get_upc=self.upc_getter,
            )

        template_path = resource_path(self.report_template_path)
        template = Template(filename=template_path)
        return template.render(**data)


class ReportOutputView(ExportMasterView):
    """
    Master view for report output
    """
    model_class = model.ReportOutput
    route_prefix = 'report_output'
    url_prefix = '/reports/generated'
    downloadable = True

    grid_columns = [
        'id',
        'report_name',
        'filename',
        'created',
        'created_by',
    ]

    form_fields = [
        'id',
        'report_name',
        'report_type',
        'params',
        'filename',
        'created',
        'created_by',
    ]

    def configure_grid(self, g):
        super(ReportOutputView, self).configure_grid(g)
        g.set_link('filename')

    def configure_form(self, f):
        super(ReportOutputView, self).configure_form(f)

        # params
        f.set_renderer('params', self.render_params)

        # filename
        if self.viewing:
            f.set_renderer('filename', self.render_download)

    def render_params(self, report, field):
        params = report.params
        if not params:
            return ""

        # TODO: should sort these, perhaps according to Report definition?
        params = [{'key': key, 'value': value}
                  for key, value in params.items()]

        route_prefix = self.get_route_prefix()
        g = grids.Grid(
            key='{}.params'.format(route_prefix),
            data=params,
            columns=['key', 'value'],
        )
        return HTML.literal(g.render_grid())

    def render_download(self, report, field):
        path = report.filepath(self.rattail_config)
        url = self.get_action_url('download', report)
        return self.render_file_field(path, url=url)

    def download(self):
        report = self.get_instance()
        path = report.filepath(self.rattail_config)
        return self.file_response(path)


class GenerateReport(View):
    """
    View for generating a new report.
    """

    def __init__(self, request):
        super(GenerateReport, self).__init__(request)
        self.handler = self.get_handler()

    def get_handler(self):
        return get_report_handler(self.rattail_config)

    def choose(self):
        """
        View which allows user to choose which type of report they wish to
        generate.
        """
        # handler is responsible for determining which report types are valid
        reports = self.handler.get_reports()

        # make form to accept user choice of report type
        schema = NewReport().bind(valid_report_types=list(reports))
        form = forms.Form(schema=schema, request=self.request)
        form.submit_label = "Continue"
        form.cancel_url = self.request.route_url('report_output')

        # TODO: should probably "group" certain reports together somehow?
        # e.g. some for customers/membership, others for product movement etc.
        values = [(r.type_key, r.name) for r in reports.values()]
        form.set_widget('report_type', forms.widgets.PlainSelectWidget(values=values,
                                                                       size=10))

        # if form validates, that means user has chosen a report type, so we
        # just redirect to the appropriate "new report" page
        if form.validate(newstyle=True):
            raise self.redirect(self.request.route_url('generate_specific_report',
                                                       type_key=form.validated['report_type']))

        return {
            'index_title': "Generate Report",
            'form': form,
            'dform': form.make_deform_form(),
            'reports': reports,
            'report_descriptions': dict([(r.type_key, r.__doc__)
                                         for r in reports.values()]),
            'use_form': self.rattail_config.getbool('tailbone', 'reporting.choosing_uses_form',
                                                    default=True),
        }

    def generate(self):
        """
        View for actually generating a new report.  Allows user to provide
        input parameters specific to the report type, then creates a new report
        and redirects user to view the output.
        """
        type_key = self.request.matchdict['type_key']
        report = self.handler.get_report(type_key)
        report_params = report.make_params(Session())

        NODE_TYPES = {
            datetime.date: colander.Date,
        }

        schema = colander.Schema()
        for param in report_params:

            # make a new node of appropriate schema type
            node_type = NODE_TYPES.get(param.type, colander.String)
            node = colander.SchemaNode(typ=node_type(), name=param.name)

            # allow empty value if param is optional
            if not param.required:
                node.missing = colander.null

            schema.add(node)

        form = forms.Form(schema=schema, request=self.request,
                          use_buefy=self.get_use_buefy())
        form.submit_label = "Generate this Report"
        form.cancel_url = self.request.route_url('generate_report')

        # must declare jquery support for date fields, ugh
        # TODO: obviously would be nice for this to be automatic?
        for param in report_params:
            if param.type is datetime.date:
                form.set_type(param.name, 'date_jquery')

        # if form validates, start generating new report output; show progress page
        if form.validate(newstyle=True):
            key = 'report_output.generate'
            progress = SessionProgress(self.request, key)
            kwargs = {'progress': progress}
            thread = Thread(target=self.generate_thread,
                            args=(report, form.validated, self.request.user.uuid),
                            kwargs=kwargs)
            thread.start()
            return self.render_progress(progress, {
                'cancel_url': self.request.route_url('report_output'),
                'cancel_msg': "Report generation was canceled",
            })

        return {
            'index_title': "Generate Report",
            'index_url': self.request.route_url('generate_report'),
            'report': report,
            'form': form,
            'dform': form.make_deform_form(),
            'use_buefy': self.get_use_buefy(),
        }

    def generate_thread(self, report, params, user_uuid, progress=None):
        """
        Generate output for the given report and params, and return the
        resulting :class:`rattail:~rattail.db.model.reports.ReportOutput`
        object.
        """
        session = RattailSession()
        user = session.query(model.User).get(user_uuid)
        try:
            output = self.handler.generate_output(session, report, params, user, progress=progress)

        # if anything goes wrong, rollback and log the error etc.
        except Exception as error:
            session.rollback()
            log.exception("Failed to generate '%s' report: %s", report.type_key, report)
            session.close()
            if progress:
                progress.session.load()
                progress.session['error'] = True
                progress.session['error_msg'] = "Failed to generate report: {}: {}".format(
                    type(error).__name__, error)
                progress.session.save()

        # if no error, check result flag (false means user canceled)
        else:
            session.commit()
            success_url = self.request.route_url('report_output.view', uuid=output.uuid)
            session.close()
            if progress:
                progress.session.load()
                progress.session['complete'] = True
                progress.session['success_url'] = success_url
                progress.session.save()

    @classmethod
    def defaults(cls, config):

        # note that we include this in the "Generated Reports" permissions group
        config.add_tailbone_permission('report_output', 'report_output.generate',
                                       "Generate new report (of any type)")

        # "generate report" (which is really "choose")
        config.add_route('generate_report', '/reports/generate')
        config.add_view(cls, attr='choose', route_name='generate_report',
                        permission='report_output.generate',
                        renderer='/reports/choose.mako')

        # "generate specific report" (accept custom params, truly generate)
        config.add_route('generate_specific_report', '/reports/generate/{type_key}')
        config.add_view(cls, attr='generate', route_name='generate_specific_report',
                        permission='report_output.generate',
                        renderer='/reports/generate.mako')


@colander.deferred
def valid_report_type(node, kw):
    valid_report_types = kw['valid_report_types']

    def validate(node, value):
        # we just need to provide possible values, and let core validator
        # handle the rest
        oneof = colander.OneOf(valid_report_types)
        return oneof(node, value)

    return validate


class NewReport(colander.Schema):

    report_type = colander.SchemaNode(colander.String(),
                                      validator=valid_report_type)


def add_routes(config):
    config.add_route('reports.ordering',        '/reports/ordering')
    config.add_route('reports.inventory',       '/reports/inventory')


def includeme(config):
    add_routes(config)

    config.add_view(OrderingWorksheet, route_name='reports.ordering',
                    renderer='/reports/ordering.mako')

    config.add_view(InventoryWorksheet, route_name='reports.inventory',
                    renderer='/reports/inventory.mako')

    # fix permission group
    config.add_tailbone_permission_group('report_output', "Generated Reports")

    # note that GenerateReport must come first, per route matching
    GenerateReport.defaults(config)
    ReportOutputView.defaults(config)
